(* Define a set of method to traverse the abstract syntax tree of Ocaml language
   ( AST ) applying a set of operation that can match some patterns in the tree.
*)

open MLast;;


(** Methods you can define to do matching in the AST *)

module type AST_ANALYZE_MATCH =
  sig
    type t
    val ctyp           : t -> ctyp -> t
    val row_field      : t -> row_field -> t
    val class_infos    : t -> 'a class_infos -> t
    val patt           : t -> patt -> t
    val expr           : t -> expr -> t
    val module_type    : t -> module_type -> t
    val sig_item       : t -> sig_item -> t
    val with_constr    : t -> with_constr -> t
    val module_expr    : t -> module_expr -> t
    val str_item       : t -> str_item -> t
    val type_decl      : t -> type_decl -> t
    val class_type     : t -> class_type -> t
    val class_sig_item : t -> class_sig_item -> t
    val class_expr     : t -> class_expr -> t
    val class_str_item : t -> class_str_item -> t
    val interf         : t -> (sig_item * loc) list -> t
    val implem         : t -> (str_item * loc) list -> t
  end
;;

module AstAnalyze = 
  functor ( AstAnalyzeMatch : AST_ANALYZE_MATCH ) ->
struct
  let if_some x f t = 
    match x with 
      Some y -> f t y
    | None -> t 

  let fold = 
    List.fold_left

  let rec ctyp t c =
    let t = 
      AstAnalyzeMatch.ctyp t c
    in
    match c with
      TyAcc (loc,c1,c2) -> 
        ctyp (ctyp t c1) c2
    | TyAli (loc,c1,c2) -> 
        ctyp (ctyp t c1) c2
    | TyAny loc -> 
        t
    | TyApp (loc,c1,c2) -> 
        ctyp (ctyp t c1) c2
    | TyArr (loc,c1,c2) -> 
        ctyp (ctyp t c1) c2
    | TyCls (loc,lst) -> 
        t
    | TyLab (loc,str,c1) -> 
        ctyp t c1
    | TyLid (loc,str) -> 
        t
    | TyMan (loc,c1,c2) -> 
        ctyp (ctyp t c1) c2
    | TyObj (loc,lst,b) -> 
        fold ( 
          fun t (str,c1) -> 
            ctyp t c1 
          ) t lst
    | TyOlb (loc,str,c1) -> 
        ctyp t c1
    | TyPol (loc,lst,c1) -> 
        ctyp t c1
    | TyQuo (loc,str) -> 
        t
    | TyRec (loc,b,lst) -> 
        fold ( 
          fun t (loc,str,b,c1) -> 
            ctyp t c1 
          ) t lst
    | TySum (loc,b,lst) -> 
        fold ( 
          fun t (loc,str,lst) -> 
            fold ctyp t lst
          ) t lst
    | TyTup (loc,lst) -> 
        fold ctyp t lst
    | TyUid (loc,str)-> 
        t
    | TyVrn (loc,lst,opt) -> 
        fold row_field t lst
  and row_field t r =
    let t =
      AstAnalyzeMatch.row_field t r
    in
    match r with
      RfTag (str,b,lst) -> 
        fold ctyp t lst 
    | RfInh c1 -> 
        ctyp t c1

  let class_infos sub_info t ci =
    let t =
      AstAnalyzeMatch.class_infos t ci
    in
    sub_info t ci.ciExp

  let rec patt t p =
    let t = 
      AstAnalyzeMatch.patt t p
    in
    match p with
      PaAcc (loc,p1,p2) ->
        patt (patt t p1) p2
    | PaAli (loc,p1,p2) ->
        patt (patt t p1) p2
    | PaAnt (loc,p1) ->
        patt t p1
    | PaAny loc ->
        t
    | PaApp (loc,p1,p2) ->
        patt (patt t p1) p2
    | PaArr (loc,lst) -> 
        fold patt t lst
    | PaChr (loc,str) -> 
        t
    | PaInt (loc,str) -> 
        t
    | PaInt32 (loc,str) -> 
        t
    | PaInt64 (loc,str) -> 
        t
    | PaNativeInt (loc,str) -> 
        t
    | PaFlo (loc,str) -> 
        t
    | PaLab (loc,str,opt) -> 
        if_some opt patt t
    | PaLid (loc,str) -> 
        t
    | PaOlb (loc,str,opt) -> 
        (
          match opt with 
            Some(p1, None) -> patt t p1
          | Some(p1, Some e1) -> expr (patt t p1) e1
          | None -> t
        )
    | PaOrp (loc,p1,p2) -> 
        patt (patt t p1) p2
    | PaRng (loc,p1,p2) -> 
        patt (patt t p1) p2
    | PaRec (loc,lst) -> 
        fold (
          fun t (p1,p2) -> 
            patt (patt t p1) p2
          ) t lst
    | PaStr (loc,str) -> 
        t
    | PaTup (loc,lst) -> 
        fold patt t lst
    | PaTyc (loc,p1,c1) -> 
        ctyp (patt t p1) c1
    | PaTyp (loc,lst) -> 
        t 
    | PaUid (loc,str) -> 
        t
    | PaVrn (loc,str) -> 
        t
  and expr t e =
    let t = 
      AstAnalyzeMatch.expr t e 
    in
    match e with
      ExAcc (loc,e1,e2) -> 
        expr (expr t e1) e2 
    | ExAnt (loc,e1) -> 
        expr t e1
    | ExApp (loc,e1,e2) -> 
        expr (expr t e1) e2 
    | ExAre (loc,e1,e2) -> 
        expr (expr t e1) e2 
    | ExArr (loc,lst) -> 
        fold expr t lst
    | ExAsf loc -> 
        t
    | ExAsr (loc,e1) -> 
        expr t e1
    | ExAss (loc,e1,e2) -> 
        expr (expr t e1) e2 
    | ExChr (loc,str) -> 
        t
    | ExCoe (loc,e1,opt,c1) -> 
        ctyp (if_some opt ctyp (expr t e1)) c1
    | ExFlo (loc,str) -> 
        t
    | ExFor (loc,str,e1,e2,b1,lst) ->
        fold expr t ( e1 :: e2 :: lst )
    | ExFun (loc,lst) ->
        fold
        ( fun t (p1,opt,e1) ->
          expr (if_some opt expr (patt t p1)) e1
        ) t lst
    | ExIfe (loc,e1,e2,e3) ->
        fold expr t [ e1 ; e2 ; e3 ]
    | ExInt (loc,str) -> 
        t
    | ExInt32 (loc,str) -> 
        t
    | ExInt64 (loc,str) -> 
        t
    | ExNativeInt (loc,str) -> 
        t
    | ExLab (loc,str,opt) -> 
        if_some opt expr t
    | ExLaz (loc,e1) -> 
        expr t e1
    | ExLet (loc,b,lst,e1) ->
        expr ( 
          fold 
          ( fun t (p1,e1) ->
            expr (patt t p1) e1
            ) t lst
        ) e1
    | ExLid (loc,str) -> 
        t
    | ExLmd (loc,str,me1,e1) ->
        expr (module_expr t me1) e1
    | ExMat (loc,e1,lst) ->
        fold 
        ( fun t (p1,opt,e1) ->
          expr (if_some opt expr (patt t p1)) e1
        ) (expr t e1) lst
    | ExNew (loc,lst) -> 
        t
    | ExObj (loc,opt,lst) ->
        fold class_str_item (if_some opt patt t) lst 
    | ExOlb (loc,str,opt) -> 
        if_some opt expr t
    | ExOvr (loc,lst) -> 
        fold
        ( fun t (str,e1) ->
          expr t e1
        ) t lst
    | ExRec (loc,lst,opt) ->
        if_some opt expr
        (
          fold
          ( fun t (p1,e1) ->
            expr (patt t p1) e1
          ) t lst
        )
    | ExSeq (loc,lst) ->
        fold expr t lst
    | ExSnd (loc,e1,str) -> 
        expr t e1
    | ExSte (loc,e1,e2) ->
        expr (expr t e1) e2
    | ExStr (loc,str) -> 
        t
    | ExTry (loc,e1,lst) ->
        fold ( 
          fun t (p1,opt,e1) ->
            expr (if_some opt expr (patt t p1)) e1
        ) (expr t e1) lst
    | ExTup (loc,lst) ->
        fold expr t lst
    | ExTyc (loc,e1,c1) ->
        ctyp (expr t e1) c1
    | ExUid (loc,str) -> 
        t
    | ExVrn (loc,str) -> 
        t
    | ExWhi (loc,e1,lst) ->
        fold expr (expr t e1) lst
  and module_type t mt =
    let t = 
      AstAnalyzeMatch.module_type t mt
    in
    match mt with
      MtAcc (loc,mt1,mt2) ->
        module_type (module_type t mt1) mt2
    | MtApp (loc,mt1,mt2) ->
        module_type (module_type t mt1) mt2
    | MtFun (loc,str,mt1,mt2) ->
        module_type (module_type t mt1) mt2
    | MtLid (loc,str) -> 
        t
    | MtQuo (loc,str) -> 
        t
    | MtSig (loc,lst) -> 
        fold sig_item t lst
    | MtUid (loc,str) -> 
        t
    | MtWit (loc,mt1,lst) ->
        fold with_constr (module_type t mt1) lst
  and sig_item t sg =
    let t =
      AstAnalyzeMatch.sig_item t sg
    in
    match sg with
      SgCls (loc,lst) -> 
        fold (class_infos class_type) t lst
    | SgClt (loc,lst) -> 
        fold (class_infos class_type) t lst
    | SgDcl (loc,lst) ->
        fold sig_item t lst
    | SgDir (loc,str,opt) ->
        if_some opt expr t
    | SgExc (loc,str,lst) ->
        fold ctyp t lst
    | SgExt (loc,str,c1,lst) ->
        ctyp t c1
    | SgInc (loc,mt1) ->
        module_type t mt1
    | SgMod (loc,str,mt1) ->
        module_type t mt1
    | SgRecMod (loc,lst) ->
        fold ( 
          fun t (str,mt1) ->
            module_type t mt1
        ) t lst
    | SgMty (loc,str,mt1) ->
        module_type t mt1
    | SgOpn (loc,lst) ->
        t
    | SgTyp (loc,lst) ->
        fold type_decl t lst
    | SgUse (loc,str,lst) ->
        fold ( 
          fun t (sg1,loc) ->
            sig_item t sg1
        ) t lst
    | SgVal (loc,str,c1) ->
        ctyp t c1
  and with_constr t wc =
    let t = 
      AstAnalyzeMatch.with_constr t wc
    in
    match wc with
      WcTyp (loc,lst1,lst2,c1) ->
        ctyp t c1      
    | WcMod (loc,lst,me1) ->
        module_expr t me1
  and module_expr t me =
    let t =
      AstAnalyzeMatch.module_expr t me
    in
    match me with
      MeAcc (loc,me1,me2) ->
        module_expr (module_expr t me1) me2
    | MeApp (loc,me1,me2) ->
        module_expr (module_expr t me1) me2
    | MeFun (loc,str,mt1,me1) ->
        module_expr (module_type t mt1) me1
    | MeStr (loc,lst) ->
        fold str_item t lst
    | MeTyc (loc,me1,mt1) ->
        module_type (module_expr t me1) mt1
    | MeUid (loc,str) ->
        t 
  and str_item t st =
    let t =
      AstAnalyzeMatch.str_item t st
    in
    match st with
      StCls (loc,lst) ->
        fold (class_infos class_expr) t lst
    | StClt (loc,lst) ->
        fold (class_infos class_type) t lst
    | StDcl (loc,lst) ->
        fold str_item t lst
    | StDir (loc,str,opt) ->
        if_some opt expr t
    | StExc (loc,str,lst,lst_str) ->
        fold ctyp t lst
    | StExp (loc,e1) ->
        expr t e1
    | StExt (loc,str,c1,lst_str) ->
        ctyp t c1
    | StInc (loc,me1) ->
        module_expr t me1
    | StMod (loc,str,me1) ->
        module_expr t me1
    | StRecMod (loc,lst) ->
        fold ( 
          fun t (str,mt1,me1) ->
            module_expr (module_type t mt1) me1
        ) t lst
    | StMty (loc,str,mt1) ->
        module_type t mt1
    | StOpn (loc,lst) ->
        t
    | StTyp (loc,lst) ->
        fold type_decl t lst
    | StUse (loc,str,lst) ->
        fold (
          fun t (st1,loc) ->
            str_item t st1
        ) t lst
    | StVal (loc,b,lst) ->
        fold ( 
          fun t (p1,e1) ->
            expr (patt t p1) e1
        ) t lst
  and type_decl t td =
    let t =
      AstAnalyzeMatch.type_decl t td
    in
    let ((loc,str),lst1,c1,lst2) = td
    in
    fold ( 
      fun t (c1,c2) ->
        ctyp (ctyp t c1) c2
    ) (ctyp t c1) lst2
  and class_type t ct =
    let t =
      AstAnalyzeMatch.class_type t ct
    in
    match ct with
      CtCon (loc,lst1,lst2) ->
        fold ctyp t lst2
    | CtFun (loc,c1,ct1) ->
        class_type (ctyp t c1) ct1
    | CtSig (loc,opt,lst) ->
        fold class_sig_item (if_some opt ctyp t) lst
  and class_sig_item t cg =
    let t =
      AstAnalyzeMatch.class_sig_item t cg
    in
    match cg with
      CgCtr (loc,c1,c2) ->
        ctyp (ctyp t c1) c2
    | CgDcl (loc,lst) ->
        fold class_sig_item t lst
    | CgInh (loc,ct1) ->
        class_type t ct1
    | CgMth (loc,str,b,c1) ->
        ctyp t c1
    | CgVal (loc,str,b,c1) ->
        ctyp t c1
    | CgVir (loc,str,b,c1) ->
        ctyp t c1
  and class_expr t ce =
    let t =
      AstAnalyzeMatch.class_expr t ce
    in
    match ce with
      CeApp (loc,ce1,e1) ->
        expr (class_expr t ce1) e1
    | CeCon (loc,lst1,lst2) ->
        fold ctyp t lst2
    | CeFun (loc,p1,ce1) ->
        class_expr (patt t p1) ce1
    | CeLet (loc,b,lst,ce1) ->
        class_expr 
        (
          fold ( 
            fun t (p1,e1) ->
              expr (patt t p1) e1
            ) t lst
        ) ce1
    | CeStr (loc,opt,lst) ->
        fold class_str_item (if_some opt patt t) lst
    | CeTyc (loc,ce1,ct1) ->
        class_type (class_expr t ce1) ct1
  and class_str_item t cr =
    let t =
      AstAnalyzeMatch.class_str_item t cr
    in
    match cr with
      CrCtr (loc,c1,c2) ->
        ctyp (ctyp t c1) c2
    | CrDcl (loc,lst) ->
        fold class_str_item t lst
    | CrInh (loc,ce1,opt) ->
        class_expr t ce1
    | CrIni (loc,e1) ->
        expr t e1
    | CrMth (loc,str,b,e1,opt) ->
        if_some opt ctyp (expr t e1)
    | CrVal (loc,str,b,e1) ->
        expr t e1
    | CrVir (loc,str,b,c1) ->
        ctyp t c1

  let interf t lst = 
    let t =
      AstAnalyzeMatch.interf t lst
    in
    fold (
      fun t (sg,loc) ->
        sig_item t sg
    ) t lst

  let implem t lst =
    let t = 
      AstAnalyzeMatch.implem t lst
    in
    fold (
      fun t (st,loc) ->
        str_item t st
    ) t lst
end
;;
